#include <stdlib.h>
#include <stdio.h>

#include "file.h"
#include "directory.h"
#include "image.h"
#include "slate.h"

extern const unsigned long int CardSize;
extern unsigned long long int currentDispatchID;

void
error (const char * message)
{
    fprintf (stderr, "Slate: %s\n", message);
    exit (EXIT_FAILURE);
}

#include "slatevm.h"

void
growMemoryBy (struct ObjectHeap * heap, unsigned long int delta)
{
    if (heap -> memoryEnd + delta > heap -> memoryLimit)
      error ("Attempted to grow heap beyond limits");
    
    heap -> memoryEnd += delta;

    fprintf (stderr, "Slate: Growing heap to %d bytes.\n", (unsigned) heap -> memoryEnd - (unsigned) heap -> memory);
}

void
shrinkMemoryBy (struct ObjectHeap * heap, unsigned long int delta)
{
    if (heap -> memoryEnd - delta <= (unsigned long int) heap -> memory)
      error ("Attempted to shrink heap to zero size");

    heap -> memoryEnd -= delta;

    fprintf (stderr, "Slate: Shrinking heap to %d bytes.\n", (unsigned) heap -> memoryEnd - (unsigned) heap -> memory);
}


ssize_t
extractCString (struct ByteArray * array, char * buffer, size_t bufferSize)
{
	ssize_t arrayLength = PSByteArray_extractInto_sized_ (array, buffer, bufferSize - 1);

	if (arrayLength < 0)
		return -1;

	buffer [arrayLength] = '\0';
		
	return arrayLength;
}


void
initModules ()
{
	initFileModule ();
	initDirectoryModule ();
}

int
slateMain (int argc, char ** argv)
{
    FILE * file;
    ImageHeader header;
    struct ObjectHeap heap;
    unsigned long int stackBottom;

    if (argc < 2)
    {
        fprintf (stderr, "Usage: %s <image>\n", argv [0]);
        return EXIT_FAILURE;
    }

    initModules ();

    file = fopen (argv [1], "rb");
    if (file == NULL)
        error ("Could not open image file");

    if (fread (& header, sizeof (ImageHeader), 1, file) != 1 ||
        header.magicNumber != MAGIC_NUMBER ||
        header.imageSize >= HEAP_LIMIT)
    {
        fclose (file);
        error ("Invalid image header");
    }

    heap.memory = (unsigned long int *) calloc (HEAP_LIMIT, 1);
    heap.pinnedCards = (unsigned long int *) calloc (HEAP_LIMIT / (CardSize * 8), 1);
    if (heap.memory == NULL ||
        heap.pinnedCards == NULL)
    {
        fclose (file);
        error ("Failed to allocate memory for image heap");
    }
            
    if (fread (heap.memory, 1, header.imageSize, file) != header.imageSize)
    {
        fclose (file);
        error ("Could not read image heap");
    }

    fclose (file);

    heap.memoryEnd = (unsigned long int) heap.memory + header.imageSize;
    heap.memoryLimit = (unsigned long int) heap.memory + HEAP_LIMIT;
    heap.lastHash = header.nextHash;
    heap.specialObjectsOop = header.specialOops;
    heap.stackBottom = & stackBottom;

    CurrentMemory = & heap;

    currentDispatchID = header.currentDispatchID;

    PSObjectHeap_initializeWithShift_ (CurrentMemory, (unsigned long int) heap.memory);
    
    PSInterpreter_interpret ((struct Interpreter *) PSObjectHeap_specialAt_ (CurrentMemory, InterpreterObject));

    return EXIT_SUCCESS;
}

int
saveImageNamed (struct ByteArray * name)
{
        char nameString [SLATE_FILE_NAME_LENGTH];
        size_t nameLength = PSObject_payloadSize ((struct Object *) name);
        FILE * file;
        ImageHeader header;

        if (nameLength >= sizeof (nameString))
          return -1;

        memcpy (nameString, name -> elements, nameLength);
        nameString [nameLength] = '\0';

        file = fopen (nameString, "wb");
        if (file == NULL)
          return -1;

       PSObjectHeap_compact_ (CurrentMemory, 0);
       
       if (CurrentMemory -> markColor == 0)
         PSObjectHeap_garbageCollect (CurrentMemory);

       PSObjectHeap_sweep (CurrentMemory);
       
       PSObjectHeap_adjustAllOopsBy_ (CurrentMemory, - (unsigned) CurrentMemory -> memory);

       header.magicNumber = MAGIC_NUMBER;
       header.imageSize = CurrentMemory -> lastAllocated - (unsigned) CurrentMemory -> memory;
       header.nextHash = CurrentMemory -> lastHash;
       header.specialOops = CurrentMemory -> specialObjectsOop - (unsigned) CurrentMemory -> memory; 
       header.currentDispatchID = currentDispatchID;

       fwrite (& header, sizeof (header), 1, file);
       fwrite (CurrentMemory -> memory, 1, header.imageSize, file);

       fclose (file);

       PSObjectHeap_adjustAllOopsBy_ (CurrentMemory, (unsigned) CurrentMemory -> memory);

       return 0;
}

